import os
from os.path import join, basename, splitext
import logging
from glob import glob
import shutil
from typing import List

from osgeo import gdal


logger = logging.getLogger(__name__)



def georeference_step(source_dir: os.PathLike, target_dir: os.PathLike, type: str, **options: dict):
    type_name = type.lower()
    if type_name == 'gcp':
        georef_func = gcp_georef
    elif type_name == 'rpc':
        georef_func = rpc_georef
    elif type_name == 'world':
        georef_func = world_georef

    for filename in glob(join(source_dir, '*.TIF')):
        target_filename = join(target_dir, basename(filename))
        georef_func(filename, target_filename, **options)


def gcp_georef(input_filename: os.PathLike, target_filename: os.PathLike, order: int=1, projection: str='EPSG:4326'):
    succeded = False

    # simple case: get the geotransform from some GCPs
    ds = gdal.Open(input_filename, gdal.GA_Update)
    if ds.GetGCPCount() <= 4:
        try:
            gcps = ds.GetGCPs()
            gt = gdal.GCPsToGeoTransform(gcps)
            ds.SetGeoTransform(gt)
        except Exception:
            del ds
            logger.warning("Failed to get geotransform via GCPsToGeoTransform")
        else:
            del ds
            shutil.move(input_filename, target_filename)
            succeded = True

    # otherwise warp
    if not succeded:
        logger.info("Applying GCP transform by warping")
        gdal.Warp(
            target_filename,
            input_filename,
            options=gdal.WarpOptions(
                polynomialOrder=order,
                dstSRS=projection,
            )
        )

def rpc_georef(input_filename: os.PathLike, target_filename: os.PathLike, rpc_file_template: str='{fileroot}.RPC', warp_options: dict=None):
    fileroot, extension = splitext(input_filename)
    rpc_file_glob = rpc_file_template.format(
        filename=input_filename, fileroot=fileroot,
        extension=extension,
    )
    try:
        rpc_filename = glob(rpc_file_glob)[0]
    except IndexError:
        logger.warn('No RPC filename found with glob %s' % rpc_file_glob)

    # rename RPC filename to be compatible with GDAL
    shutil.move(rpc_filename, '%s.rpc' % input_filename)

    gdal.Warp(
        target_filename,
        input_filename,
        rpc=True,
        **(warp_options or {})
    )

def corner_georef(input_filename: os.PathLike, target_filename: os.PathLike, corner_names: List[str]=None):
    corner_names = corner_names or ["bottom_left", "bottom_right", "top_left", "top_right"]
    ds = gdal.Open(input_filename, gdal.GA_Update)
    for corner_name in corner_names:
        ds.GetMetaData()


def world_georef():
    pass




def gcps_from_borders(dst, coords, orbit_direction, force_north_up=False):
    # expects coordinates in dict(.*border_left.*:[lat,lon],...)
    gcps = []
    if force_north_up and len(coords) == 4:
        # compute gcps assuming north-up, east-right image no matter, what is claimed by metadata
        sorted_by_lats = sorted(coords.values(), key=lambda x: x[0], reverse=True)
        # compare longitudes
        if sorted_by_lats[0][1] > sorted_by_lats[1][1]:
            #                                                            /\
            #  1                                                         \ \
            # top - left corner has lower latitude from two northernmost  \/
            if orbit_direction != "descending":
                top_left = sorted_by_lats[1]
                top_right = sorted_by_lats[0]
                bottom_left = sorted_by_lats[3]
                bottom_right = sorted_by_lats[2]
            else:
                top_left = sorted_by_lats[3]
                top_right = sorted_by_lats[2]
                bottom_left = sorted_by_lats[1]
                bottom_right = sorted_by_lats[0]
        else:
            #                                                                /\
            #   2                                                           / /
            #  top - left corner has higher latitude from two northernmost  \/
            if orbit_direction != "descending":
                top_left = sorted_by_lats[0]
                top_right = sorted_by_lats[1]
                bottom_left = sorted_by_lats[2]
                bottom_right = sorted_by_lats[3]
            else:
                top_left = sorted_by_lats[2]
                top_right = sorted_by_lats[3]
                bottom_left = sorted_by_lats[0]
                bottom_right = sorted_by_lats[1]
        gcps.append(gdal.GCP(bottom_left[1], bottom_left[0], 0, 0.5, 0.5))
        gcps.append(gdal.GCP(bottom_right[1], bottom_right[0], 0, dst.RasterXSize - 0.5, 0.5))
        gcps.append(gdal.GCP(top_left[1], top_left[0], 0, 0.5, dst.RasterYSize - 0.5))
        gcps.append(gdal.GCP(top_right[1], top_right[0], 0, dst.RasterXSize - 0.5, dst.RasterYSize - 0.5))
    else:
        for key, value in coords.items():
            # assume points are labeled correctly and image not necessarily north up, access border points as mid-pixels and use them as GCPs
            if "bottom_left" in key.lower() or "bottom_right" in key.lower():
                if orbit_direction != "descending":
                    y = 0.5
                else:
                    y = dst.RasterYSize - 0.5
            else:
                if orbit_direction != "descending":
                    y = dst.RasterYSize - 0.5
                else:
                    y = 0.5
            if "bottom_left" in key.lower() or "top_left" in key.lower():
                if orbit_direction != "descending":
                    x = 0.5
                else:
                    x = dst.RasterXSize - 0.5
            else:
                if orbit_direction != "descending":
                    x = dst.RasterXSize - 0.5
                else:
                    x = 0.5
            gcps.append(gdal.GCP(value[1], value[0], 0, x, y))
    return gcps
